(***********************************************************************)
(*                                                                     *)
(*                                Sklml                                *)
(*                                                                     *)
(*                 The OCaml parallel skeleton system                  *)
(*                                                                     *)
(*                       EPI Serena, INRIA Paris                       *)
(*                                                                     *)
(*  Copyright (c) 2007-2017 INRIA.                                     *)
(*  Institut National de Recherche en Informatique et en Automatique.  *)
(*  All rights reserved. This file is distributed only by permission.  *)
(*                                                                     *)
(*  Quentin Carbonneaux                                                *)
(*  Francois Clement <Francois.Clement@inria.fr>                       *)
(*  Pierre Weis <Pierre.Weis@inria.fr>                                 *)
(*                                                                     *)
(*  This example is based on an example designed for OCamlP3l.         *)
(*                                                                     *)
(***********************************************************************)

(* $Id: gnuplot.ml,v 1.6 2017/01/04 11:58:30 fclement Exp $ *)

module Nat = Unatural;;

open Rdec2;;
open Dd_config;;

(* The gnuplot version number is set by HTMLC from the knowledge got by an
   external shell script. *)
let gnuplot_version_number = 4.0;;

let check_graphical_output = function
  | No -> ()
  | Postscript
  | Window as go ->
    if gnuplot_version_number <= 0. then begin
      prerr_endline
        (Printf.sprintf
           "Graphical output %s needs Gnuplot."
           (string_of_graphical_output go));
      exit 4
    end
;;

(* The syntax of some gnuplot commands depends on the version number *)
let (unset, pointsize) =
  let (us, ps) =
    if gnuplot_version_number >= 4.0 then
      ((fun s -> Printf.sprintf "\nunset %s" s),
       (fun () -> Printf.sprintf ""))
    else
      ((fun s -> Printf.sprintf "\nset no%s" s),
       (fun () -> Printf.sprintf "pointsize 0.5")) in
  ((fun () -> us), (fun () -> ps))
;;

(* Generating the gnuplot command *)

let make_isosample () (nsbdx, nsbdy) =
  Printf.sprintf "set isosample %d, %d"
    (Nat.to_int (Nat.succ nsbdx))
    (Nat.to_int (Nat.succ nsbdy))

and make_term_or_title () (output, res, iter) =
  let iter = Nat.to_int iter in
  match output with
  | No
  | Postscript ->
    Printf.sprintf
      "%a%a%a%a\
      \nset terminal postscript eps color solid\
      \nset output '%s%d.eps'"
      unset "xtics"
      unset "ytics"
      unset "ztics"
      unset "border"
      res iter
  | Window -> Printf.sprintf "\nset title 'iter = %d'" iter

and make_splot () (res, nsbd) =
  let s = ref "\nsplot [0:1][0:1][0:0.07]" in
  let to_do n =
    let resn = Printf.sprintf "%s%d" res (Nat.to_int n) in
    if not (Sys.file_exists resn) then begin
      Printf.eprintf "File %s not found!\n" resn; flush stderr
    end;
    let file = Printf.sprintf " '%s'" resn in
    let style =
      Printf.sprintf " notitle with points pointtype 7%a," pointsize () in
    s := !s ^ file ^ style in
  for_all_subdomains nsbd to_do;
  !s

and make_reference_surface () output =
  let f = "x*(1-x)*y*(1-y)"
  and wl = "with lines" in
  match output with
  | No
  | Window -> Printf.sprintf " %s %s\n" f wl
  | Postscript -> Printf.sprintf " %s notitle %s linetype 9 linewidth 4\n" f wl
;;

let make_plot_command output res iter nsbd =
  Printf.sprintf "%a%a%a%a"
    make_isosample nsbd
    make_term_or_title (output, res, iter)
    make_splot (res, nsbd)
    make_reference_surface output
;;
