(***********************************************************************)
(*                                                                     *)
(*                                Sklml                                *)
(*                                                                     *)
(*                 The OCaml parallel skeleton system                  *)
(*                                                                     *)
(*                       EPI Serena, INRIA Paris                       *)
(*                                                                     *)
(*  Copyright (c) 2007-2017 INRIA.                                     *)
(*  Institut National de Recherche en Informatique et en Automatique.  *)
(*  All rights reserved. This file is distributed only by permission.  *)
(*                                                                     *)
(*  Quentin Carbonneaux                                                *)
(*  Francois Clement <Francois.Clement@inria.fr>                       *)
(*  Pierre Weis <Pierre.Weis@inria.fr>                                 *)
(*                                                                     *)
(***********************************************************************)

(* $Id: parser_char.ml,v 1.15 2017/01/04 11:58:31 fclement Exp $ *)

module Nat = Unatural;;
module L = Ulist;;
module S = Ustring;;

(* In the following, we assume that n <= S.length s.
   Therefore, the next Nat.sub will not fail. *)
let string_skipn n s = S.sub s n (Nat.sub (S.length s) n);;

let _0_ = Nat._0_
and _1_ = Nat._1_
;;

open Parser;;

let pchar c =
  let parse s =
    if S.length s = _0_ then None else
    (* Now, we have S.length s > 0.
       Therefore, the next S.unsafe_get will be safe and the next
       string_skipn will not fail. *)
    let s0 = S.unsafe_get s _0_ in
    if s0 = c then Some (s0, string_skipn _1_ s) else
    None in
  { parse; }
;;

let pchar_ign c = liftM ignore (pchar c);;

let pclass l = any (L.map pchar l);;

let pnchar c =
  let parse s =
    if S.length s = _0_ then None else
    (* Now, we have S.length s > 0.
       Therefore, the next S.unsafe_get will be safe and the next
       string_skipn will not fail. *)
    let s0 = S.unsafe_get s _0_ in
    if s0 <> c then Some (s0, string_skipn _1_ s) else
    None in
  { parse; }
;;

let pword w =
  let l = S.length w in
  let rec loop i =
    if i >= l then return () else
    (* Now, we have i < l = S.length s.
       Therefore, the next S.unsafe_get will be safe. *)
    pchar (S.unsafe_get w i) >> loop (Nat.succ i) in
  loop _0_
;;

let pspaces = liftM ignore (pclass [ ' '; '\r'; '\n'; '\t'; ]);;

let pspaced p = many pspaces >> p;;

let pint =
  let pnum1 =
    let digits = [ '0'; '1'; '2'; '3'; '4'; '5'; '6'; '7'; '8'; '9'; ] in
    let int_of_char c = Char.code c - Char.code '0' in
    liftM int_of_char (pclass digits) in
  let rec loop n =
    (pnum1 >>= fun d -> loop (n * 10 + d)) +++ (return n) in
  maybe (pchar_ign '-') >>= function
    | None -> pnum1 >>= loop
    | Some () -> liftM (fun x -> -x) (pnum1 >>= loop)
;;

let pfloat =
  let pnum1 =
    let digits = [ '0'; '1'; '2'; '3'; '4'; '5'; '6'; '7'; '8'; '9'; ] in
    let int_of_char c = Char.code c - Char.code '0' in
    liftM int_of_char (pclass digits) in
  let rec loop n =
    (pnum1 >>= fun d -> loop (n *. 10. +. float_of_int d)) +++
    (pchar_ign '.' >> (pmantissa 10. 0. >>= fun s -> return (n +. s))) +++
    (return n)
  and pmantissa k s =
    (pnum1 >>= fun d -> pmantissa (k *. 10.) (float_of_int d /. k +. s)) +++
    (return s) in
  maybe (pchar_ign '-') >>= function
    | None -> liftM float_of_int pnum1 >>= loop
    | Some () -> liftM (fun x -> -. x) (liftM float_of_int pnum1 >>= loop)
;;
