(***********************************************************************)
(*                                                                     *)
(*                                Sklml                                *)
(*                                                                     *)
(*                 The OCaml parallel skeleton system                  *)
(*                                                                     *)
(*                       EPI Serena, INRIA Paris                       *)
(*                                                                     *)
(*  Copyright (c) 2007-2017 INRIA.                                     *)
(*  Institut National de Recherche en Informatique et en Automatique.  *)
(*  All rights reserved. This file is distributed only by permission.  *)
(*                                                                     *)
(*  Quentin Carbonneaux                                                *)
(*  Francois Clement <Francois.Clement@inria.fr>                       *)
(*  Pierre Weis <Pierre.Weis@inria.fr>                                 *)
(*                                                                     *)
(***********************************************************************)

(* $Id: darray.ml,v 1.23 2017/01/04 11:58:27 fclement Exp $ *)

(* Dynamic arrays. *)

module Nat = Unatural;;
module Perv = Upervasives;;
module A = Uarray;;

let _0_ = Nat._0_
and _2_ = Nat._2_
and _16_ = Nat.unsafe_of_int 16
;;

let nat_sys_max_array_length = Nat.unsafe_of_int Sys.max_array_length;;

open Tools_common_types;;

let qual = Perv.qualify "Darray";;
let invalid_arg = Perv.invalid_arg_qual qual;;

type 'a data = 'a option array

(* Invariant length < size = A.length ('a data). *)
type 'a t = (size * length * 'a data) ref
and 'a darray = 'a t
;;

exception Out_of_memory;;

let min_size = _16_;;

let increase i = if i = _0_ then min_size else Nat.mul _2_ i;;

let realloc da =
  let (sz, len, a) = !da in
  if len < sz then () else
  let sz' = increase sz in
  if sz' > nat_sys_max_array_length then raise Out_of_memory else
  let a' = A.make sz' None in
  A.blit a _0_ a' _0_ sz;
  da := (sz', len, a')
;;

let create sz =
  if sz = _0_ then invalid_arg "create" "zero size" else
  let sz = max sz min_size in
  ref (sz, _0_, A.make sz None)
;;

let length da =
  let (_, len, _) = !da in
  len
;;

let append da x =
  realloc da;
  let (sz, len, a) = !da in
  (* Now, we have len < sz = A.length a.
     Therefore, the next A.unsafe_set will be safe. *)
  A.unsafe_set a len (Some x);
  da := (sz, Nat.succ len, a);
  len
;;

let get da i =
  let (_, len, a) = !da in
  if i >= len then invalid_arg "get" "index out of bounds" else
  (* Now, we have i < len <= A.length a.
     Therefore, the next Nat.unsafe_get will be safe. *)
  match A.unsafe_get a i with
  | None -> assert false
  | Some x -> x
;;

let set da i x =
  let (_, len, a) = !da in
  if i >= len then invalid_arg "set" "index out of bounds" else
  (* Now, we have i < len <= A.length a.
     Therefore, the next Nat.unsafe_set will be safe. *)
  A.unsafe_set a i (Some x);
;;

let fold_left f accu da =
  let (_, _, a) = !da in
  let ff accu = function
    | None -> accu
    | Some x -> f accu x in
  A.fold_left ff accu a
;;
