(***********************************************************************)
(*                                                                     *)
(*                                Sklml                                *)
(*                                                                     *)
(*                 The OCaml parallel skeleton system                  *)
(*                                                                     *)
(*                       EPI Serena, INRIA Paris                       *)
(*                                                                     *)
(*  Copyright (c) 2007-2017 INRIA.                                     *)
(*  Institut National de Recherche en Informatique et en Automatique.  *)
(*  All rights reserved. This file is distributed only by permission.  *)
(*                                                                     *)
(*  Quentin Carbonneaux                                                *)
(*  Francois Clement <Francois.Clement@inria.fr>                       *)
(*  Pierre Weis <Pierre.Weis@inria.fr>                                 *)
(*                                                                     *)
(***********************************************************************)

(* $Id: sklstream.ml,v 1.65 2017/01/04 11:58:27 fclement Exp $ *)

module Nat = Unatural;;
module L = Ulist;;
module A = Uarray;;

let _0_ = Nat._0_;;

(* This is the core of the stream feature for Sklml. *)
module type SOW =
sig
  type 'a wire_data = (* private *)
    | Begin_sub_stream
    | Data of 'a
    | End_sub_stream
  and 'a cell = private
    | Nil
    | Cons of 'a wire_data * 'a t
  and 'a t = 'a cell Lazy.t
  ;;

  val nil_cell : 'a cell;;
  val cons_wire_data_cell : 'a wire_data * 'a t -> 'a cell;;
  val cons_data_cell : 'a * 'a t -> 'a cell;;
  val cons_begin_cell : 'a t -> 'a cell;;
  val cons_end_cell : 'a t -> 'a cell;;
  val get_cell : 'a t -> 'a cell;;
  val nil : 'a t;;
  val cons : 'a * 'a t -> 'a t;;
  val cons_sub_stream : 'a t * 'a t -> 'a cell;;
end
;;

module Stream_on_wire : SOW =
struct
  type 'a wire_data =
    | Begin_sub_stream
    | Data of 'a
    | End_sub_stream
  and 'a cell =
    | Nil
    | Cons of 'a wire_data * 'a t
  and 'a t = 'a cell Lazy.t
  ;;

  let nil_cell = Nil
  and cons_begin_cell s = Cons (Begin_sub_stream, s)
  and cons_data_cell (x, s) = Cons (Data x, s)
  and cons_end_cell s = Cons (End_sub_stream, s)
  and cons_wire_data_cell (d, s) = Cons (d, s)
  ;;

  let get_cell = Lazy.force;;

  (* Stream construction functions. *)
  let nil = lazy Nil;;
  let cons (x, s) = lazy (cons_data_cell (x, s));;

  let cons_sub_stream (sub, s) =
    let rec loop sub =
      match get_cell sub with
      | Nil -> cons_end_cell s
      | Cons (d, sub) -> cons_wire_data_cell (d, lazy (loop sub)) in
    cons_begin_cell (lazy (loop sub))
  ;;
end
;;

open Stream_on_wire;;

type 'a t = 'a Stream_on_wire.t
and 'a stream = 'a t
;;

(* Generate the lazy stream from a function. *)
let rec of_fun =
  let loop f =
    match f () with
    | None -> nil
    | Some x -> cons (x, of_fun f) in
  fun f -> lazy (get_cell (loop f))
;;

(* Generate a lazy stream from an input channel. *)
let of_channel input ic =
  let get_next_element () =
    try Some (input ic) with
    | End_of_file -> None in
  of_fun get_next_element
;;

let of_channel_by_lines ic = of_channel input_line ic;;
let of_channel_by_chars ic = of_channel input_char ic;;

(* Generate the lazy stream from a list. *)
let rec of_list =
  let in_of_list = function
    | [] -> nil
    | x :: l -> cons (x, of_list l) in
  fun l -> lazy (get_cell (in_of_list l))
;;

(* Generate the lazy stream from a vector. *)
let of_vect v = of_list (A.to_list v);;

(* Generate the function successively giving the elements of a lazy stream. *)
let to_fun s =
  let s0 = ref s in
  let rec loop () =
    match get_cell !s0 with
    | Nil -> None
    | Cons (Begin_sub_stream, t)
    | Cons (End_sub_stream, t) -> s0 := t; loop ()
    | Cons (Data x, t) -> s0 := t; Some x in
  loop
;;

(* Generate the list of elements of a lazy stream.
   Warning: this function is eager, meaning that the stream argument is
   completely evaluated (if this is possible, otherwise the function loops for
   ever). *)
let to_list s =
  let rec loop accu t =
    match get_cell t with
    | Nil -> L.rev accu
    | Cons (Begin_sub_stream, t)
    | Cons (End_sub_stream, t) -> loop accu t
    | Cons (Data x, t) -> loop (x :: accu) t in
  loop [] s
;;

(* Generate the vector of elements of a lazy stream.
   Warning: this function is eager, meaning that the stream argument is
   completely evaluated (if this is possible, otherwise the function loops
   for ever). *)
let to_vect s = A.of_list (to_list s);;

(* Vector and list serialization. *)
let serialize_list (l, s) = cons_sub_stream (of_list l, s);;
let serialize_vector (v, s) = serialize_list (A.to_list v, s);;

let unserialize_list s =
  let rec loop accu s =
    match get_cell s with
    | Nil
    | Cons (Begin_sub_stream, _) -> assert false
    | Cons (Data x, t) -> loop (x :: accu) t
    | Cons (End_sub_stream, t) -> (L.rev accu, t) in
  loop [] s
;;

let unserialize_vector s =
  let (l, s) = unserialize_list s in
  (A.of_list l, s)
;;

let rec serialize_stream_vector =
  let rec loop s =
    match get_cell s with
    | Nil -> nil_cell
    | Cons (Begin_sub_stream, t) -> cons_begin_cell (lazy (end_critical t))
    | Cons (Data v, t) -> serialize_vector (v, serialize_stream_vector t)
    | Cons (End_sub_stream, t) -> cons_end_cell (serialize_stream_vector t)
  and end_critical s =
    match get_cell s with
    | Nil -> assert false
    | Cons ((Begin_sub_stream | Data _), _) -> loop s
    | Cons (End_sub_stream, _) ->
      serialize_list ([], serialize_stream_vector s) in
  fun s -> lazy (loop s)
;;
let of_sarray = serialize_stream_vector;;

let unserialize_stream_vector =
  let rec loop s =
    match get_cell s with
    | Nil -> nil_cell
    | Cons (Begin_sub_stream, t) -> after_begin t
    | Cons ((Data _ | End_sub_stream), _) -> assert false
  and after_begin s =
    match get_cell s with
    | Nil -> assert false
    | Cons (Begin_sub_stream, t) -> cons_begin_cell (lazy (after_begin t))
    | Cons ((Data _ | End_sub_stream), _) ->
      let (v, t) = unserialize_vector s in
      cons_data_cell (v, lazy (after_end t))
  and after_end s =
    match get_cell s with
    | Nil -> nil_cell
    | Cons (Begin_sub_stream, t) -> after_begin t
    | Cons (Data _, _) -> assert false
    | Cons (End_sub_stream, t) -> cons_end_cell (lazy (after_end t)) in
  fun s -> lazy (loop s)
;;
let to_sarray = unserialize_stream_vector;;

let rec map f =
  let in_map s =
    match get_cell s with
    | Nil -> nil_cell
    | Cons (Begin_sub_stream, t) -> cons_begin_cell (map f t)
    | Cons (Data x, t) -> cons_data_cell (f x, map f t)
    | Cons (End_sub_stream, t) -> cons_end_cell (map f t) in
  fun s -> lazy (in_map s)
;;
let lift = map;;

let fold_left f z =
  let rec loop s =
    match get_cell s with
    | Nil -> nil_cell
    | Cons (Begin_sub_stream, t) -> after_begin t
    | Cons (Data x, t) -> fold_simpl (f z x) t
    | Cons (End_sub_stream, _) -> assert false
  and fold_simpl z s =
    match get_cell s with
    | Nil -> cons_data_cell (z, nil)
    | Cons (Begin_sub_stream, _) -> assert false
    | Cons (Data x, t) -> fold_simpl (f z x) t
    | Cons (End_sub_stream, _) -> assert false
  and after_begin s =
    match get_cell s with
    | Nil -> assert false
    | Cons (Begin_sub_stream, t) -> cons_begin_cell (lazy (after_begin t))
    | Cons ((Data _ | End_sub_stream), _) -> accu_fold_left z s
  and accu_fold_left z s =
    match get_cell s with
    | Nil -> assert false
    | Cons (Begin_sub_stream, _) -> assert false
    | Cons (Data x, t) -> accu_fold_left (f z x) t
    | Cons (End_sub_stream, t) -> cons_data_cell (z, lazy (after_end t))
  and after_end s =
    match get_cell s with
    | Nil -> nil_cell
    | Cons (Begin_sub_stream, t) -> after_begin t
    | Cons (Data _, _) -> assert false
    | Cons (End_sub_stream, t) -> cons_end_cell (lazy (after_end t)) in
  fun s -> lazy (loop s)
;;

let singleton x = cons (x, nil);;

let head s =
  match get_cell s with
  | Nil -> failwith "head: premature end of stream"
  | Cons (Begin_sub_stream, _) ->
    failwith "head: unexpected beginning of sub stream"
  | Cons (Data x, _) -> x
  | Cons (End_sub_stream, _) -> failwith "head: unexpected end of sub stream"
;;

(* Given a stream starting with a (presumably single) sub stream having a
  (presumably single element), [head_sub_stream] returns this element.
  See [unlift] below for details. *)
let head_sub_stream =
  let rec loop s =
    match get_cell s with
    | Nil ->
      failwith
        "head: premature end of stream while looking for \
         beginning of sub stream"
    | Cons (Begin_sub_stream, t) -> head_sub_stream_elem t
    | Cons (Data _, _) ->
      failwith
        "head: premature element while looking for beginning of sub stream"
    | Cons (End_sub_stream, _) ->
      failwith
        "head: premature end of sub stream while looking for \
         beginning of sub stream"
  and head_sub_stream_elem t =
    match get_cell t with
    | Nil ->
      failwith "head: premature end of stream  while looking for element"
    | Cons (Begin_sub_stream, _) ->
      failwith "head: found begin of sub stream while looking for element"
    | Cons (Data x, t) -> head_sub_stream_end x t
    | Cons (End_sub_stream, _) ->
      failwith "head: found end of sub stream while looking for element"
  and head_sub_stream_end x t =
    match get_cell t with
    | Nil ->
      failwith
        "head: premature end of stream while looking for end of sub stream"
    | Cons (Begin_sub_stream, _) ->
      failwith
        "head: found begin of sub stream while looking for end of sub stream"
    | Cons (Data _, _) ->
      failwith "head: found a data while looking for end of sub stream"
    | Cons (End_sub_stream, _) -> x in
  fun s -> loop s
;;

let unlift skl = fun x ->
  head_sub_stream (skl (lazy (cons_sub_stream (singleton x, nil))))
;;

let rec split = fun s ->
  let split_cell s =
    match get_cell s with
    | Nil -> (nil_cell, nil_cell)
    | Cons (Begin_sub_stream, t) ->
      (cons_begin_cell (fst (split t)),
       cons_begin_cell (snd (split t)))
    | Cons (Data (x, y), t) ->
      (cons_data_cell (x, fst (split t)),
       cons_data_cell (y, snd (split t)))
    | Cons (End_sub_stream, t) ->
      (cons_end_cell (fst (split t)),
       cons_end_cell (snd (split t))) in
  (lazy (fst (split_cell s)),
   lazy (snd (split_cell s)))
;;

let rec combine = fun s1 s2 ->
  let in_combine s1 s2 =
    match (get_cell s1, get_cell s2) with
    | (Nil, Nil) -> nil_cell
    | (Cons (Begin_sub_stream, t1), Cons (Begin_sub_stream, t2)) ->
      cons_begin_cell (combine t1 t2)
    | (Cons (Data x1, t1), Cons (Data x2, t2)) ->
      cons_data_cell ((x1, x2), combine t1 t2)
    | (Cons (End_sub_stream, t1), Cons (End_sub_stream, t2)) ->
      cons_end_cell (combine t1 t2)
    | ((Cons ((Begin_sub_stream | Data _ | End_sub_stream), _) | Nil),
       (Cons ((Begin_sub_stream | Data _ | End_sub_stream), _) | Nil)) ->
      assert false in
  lazy (in_combine s1 s2)
;;

let in_prod (f, g) = fun s ->
  let h (x1, x2) = (f x1, g x2) in
  let combine (s1, s2) = combine s1 s2 in
  combine (h (split s))
;;

type ('a, 'b) sum =
  | Inl of 'a
  | Inr of 'b
;;

type control =
  | C_left
  | C_right
;;

let rec sum_split = fun s ->
  let split_cell s =
    match get_cell s with
    | Nil -> ((nil, nil), nil)
    | Cons (Begin_sub_stream, t) ->
      let ((sa, sb), sc) = sum_split t in
      ((lazy (cons_begin_cell sa), lazy (cons_begin_cell sb)),
       lazy (cons_begin_cell sc))
    | Cons (Data (Inl x), t) ->
      let ((sa, sb), sc) = sum_split t in
      ((lazy (cons_data_cell (x, sa)), sb),
       lazy (cons_data_cell (C_left, sc)))
    | Cons (Data (Inr x), t) ->
      let ((sa, sb), sc) = sum_split t in
      ((sa, lazy (cons_data_cell (x, sb))),
       lazy (cons_data_cell (C_right, sc)))
    | Cons (End_sub_stream, t) ->
      let ((sa, sb), sc) = sum_split t in
      ((lazy (cons_end_cell sa), lazy (cons_end_cell sb)),
       lazy (cons_end_cell sc)) in
  ((lazy (get_cell (fst (fst (split_cell s)))),
    lazy (get_cell (snd (fst (split_cell s))))),
   lazy (get_cell (snd (split_cell s))))
;;

let rec sum_combine = fun ((sa, sb), sc) ->
  let in_combine sa sb sc =
    match get_cell sc with
    | Nil ->
      begin match (get_cell sa, get_cell sb) with
      | (Nil, Nil) -> nil_cell
      | ((Cons _ | Nil), (Cons _ | Nil)) -> assert false
      end
    | Cons (Begin_sub_stream | End_sub_stream as h, t) ->
      let (ta, tb) =
        match (get_cell sa, get_cell sb) with
        | (Cons (ha, ta), Cons (hb, tb)) ->
          assert (h = ha && h = hb); (ta, tb)
        | ((Cons _ | Nil), (Cons _ | Nil))  ->
          assert false in
      cons_wire_data_cell (h, sum_combine ((ta, tb), t))
    | Cons (Data C_left, t) ->
      let (dat, ta) =
        match get_cell sa with
        | Nil
        | Cons ((Begin_sub_stream | End_sub_stream), _) -> assert false
        | Cons (Data _ as dat, t) -> (dat, t) in
      cons_wire_data_cell (dat, sum_combine ((ta, sb), t))
    | Cons (Data C_right, t) ->
      let (dat, tb) =
        match get_cell sb with
        | Nil
        | Cons ((Begin_sub_stream | End_sub_stream), _) -> assert false
        | Cons (Data _ as dat, t) -> (dat, t) in
      cons_wire_data_cell (dat, sum_combine ((sa, tb), t)) in
  lazy (in_combine sa sb sc)
;;

let in_sum (f, g) = fun s ->
  let h ((a, b), c) = ((f a, g b), c) in
  sum_combine (h (sum_split s))
;;

let fork = fun s ->
  let rec loop s =
    match get_cell s with
    | Nil -> (nil, nil)
    | Cons (x, t) ->
      let pair = lazy (loop t) in
      (lazy (cons_wire_data_cell
         (x, lazy (get_cell (snd (Lazy.force pair))))),
       lazy (get_cell (fst (Lazy.force pair)))) in
  let rec double_control s =
    match get_cell s with
    | Nil -> nil_cell
    | Cons (Begin_sub_stream | End_sub_stream as h, t) ->
      cons_wire_data_cell
        (h, lazy (cons_wire_data_cell (h, lazy (double_control t))))
    | Cons (Data _ as h, t) ->
      cons_wire_data_cell (h, lazy (double_control t)) in
  let pair = lazy (loop (lazy (double_control s))) in
  (lazy (get_cell (fst (Lazy.force pair))),
   lazy (get_cell (snd (Lazy.force pair))))
;;

(* Merges two lazy stream into a single lazy stream with intermixed elements
   from both streams in a fair way (alternatively from one stream at a time).
   Note that the two streams must not have the same number of elements: in
   case one stream is exhausted, the result is the other (the longer one). *)
let join =
  let rec loop (s1, s2) =
    match get_cell s1 with
    | Nil -> get_cell s2
    | Cons (x, t) -> cons_wire_data_cell (x, lazy (loop (s2, t))) in
  let rec single_control s =
    match get_cell s with
    | Nil -> nil_cell
    | Cons ((Begin_sub_stream | End_sub_stream), t) -> send_control t
    | Cons (Data _ as h, t) ->
      cons_wire_data_cell (h, lazy (single_control t))
  and send_control s =
    match get_cell s with
    | Nil
    | Cons (Data _, _) -> assert false
    | Cons (Begin_sub_stream | End_sub_stream as h, t) ->
      cons_wire_data_cell (h, lazy (single_control t)) in
  fun (s1, s2) -> lazy (single_control (lazy (loop (s1, s2))))
;;

let in_par (f1, f2) =
  let f (x1, x2) = (f1 x1, f2 x2) in
  fun is -> join (f (fork is))
;;

(* [iter f s] iterates the function [f] on the stream [s]:
   if [s] is [x1; x2; ...; xn], then [iter f s] performs the sequence
   [f x1; f x2; ...; f xn].
   Note: any call to [iter] completely unfreeezes its the stream argument. *)
let rec iter f s =
  match get_cell s with
  | Nil -> ()
  | Cons ((Begin_sub_stream | End_sub_stream), t) -> iter f t
  | Cons (Data x, t) -> f x; iter f t
;;

(* Loop functions *)

let rec filter p =
  let loop s1 s2 =
    match (get_cell s1, get_cell s2) with
    | (Nil, Nil) -> nil_cell
    | (Cons (Begin_sub_stream, t1), Cons (Begin_sub_stream, t2)) ->
      cons_begin_cell (filter p t1 t2)
    | (Cons (Data x, t1), Cons (Data y, t2)) ->
      if p x then cons_data_cell (y, filter p t1 t2) else
      get_cell (filter p t1 t2)
    | (Cons (End_sub_stream, t1), Cons (End_sub_stream, t2)) ->
      cons_end_cell (filter p t1 t2)
    | ((Nil | Cons ((Begin_sub_stream | Data _ | End_sub_stream), _)),
       (Nil | Cons ((Begin_sub_stream | Data _ | End_sub_stream), _))) ->
      assert false in
  fun sb s -> lazy (loop sb s)
;;

let with_ht s def func =
  match get_cell s with
  | Nil -> def
  | Cons (hd, tl) -> func hd tl
;;

let rec get_fst_data s =
  with_ht s (fun _ -> nil_cell)
    (fun hd tl ->
      match hd with
      | Begin_sub_stream ->
        fun k -> cons_begin_cell (lazy ((get_fst_data tl) k))
      | Data x -> fun k -> cons_data_cell (x, lazy (k tl))
      | End_sub_stream -> fun k -> cons_end_cell (lazy ((get_fst_data tl) k)))
;;

let loop_core lb ls1 s2 =
  let rec loop sb s1 s2 =
    match (get_cell sb, get_cell s1) with
    | (Cons (Begin_sub_stream, tb), Cons (Begin_sub_stream, t1))
    | (Cons (End_sub_stream, tb), Cons (End_sub_stream, t1)) -> loop tb t1 s2
    | (Cons (Data b, tb), Cons (Data x1, t1)) ->
      if b then cons_data_cell (x1, lazy (loop tb t1 s2)) else
      get_fst_data s2 (loop tb t1)
    | ((Nil | Cons ((Begin_sub_stream | Data _ | End_sub_stream), _)),
       (Nil | Cons ((Begin_sub_stream | Data _ | End_sub_stream), _))) ->
      assert false in
  lazy (get_fst_data s2 (loop (Lazy.force lb) (Lazy.force ls1)))
;;

let loop p f is =
  let rec los1 = lazy (f (Lazy.force lis1))
  and los2 = lazy (p (Lazy.force lis2))
  and lis1 = lazy (loop_core los2 los1 is)
  and lis2 = lazy (Lazy.force los1) in
  lazy (get_cell (filter not (Lazy.force los2) (Lazy.force los1)))
;;

(* Rails functions *)
let fork_rails n s =
  let rec loop i s =
    match get_cell s with
    | Nil -> A.make n nil
    | Cons (Begin_sub_stream | End_sub_stream as bg, t) ->
      let arr = lazy (loop _0_ t) in
      A.init n (fun j ->
        lazy (cons_wire_data_cell
          (bg, lazy (get_cell (A.get (Lazy.force arr) j)))))
    | Cons (Data _ as x, t) ->
      let arr = lazy (loop (Nat.rem (Nat.succ i) n) t) in
      A.init n (fun j ->
        if i <> j then lazy (get_cell (A.get (Lazy.force arr) j)) else
        lazy (cons_wire_data_cell
          (x, lazy (get_cell (A.get (Lazy.force arr) j))))) in
  let arr = lazy (loop _0_ s) in
  A.init n (fun i -> lazy (get_cell (A.get (Lazy.force arr) i)))
;;

let ignore_tag bg s =
  match get_cell s with
  | Nil -> assert false
  | Cons (x, t) -> if x = bg then t else assert false
;;

let ignore_nil s =
  match get_cell s with
  | Nil -> ()
  | Cons _ -> assert false
;;

let join_rails sv =
  let n = A.length sv in
  (* Preserves the argument *)
  let sv = A.copy sv in
  let rec loop i =
    match get_cell (A.get sv i) with
    | Nil ->
      A.iter ignore_nil sv;
      nil_cell
    | Cons (Begin_sub_stream | End_sub_stream as bg, _) ->
      (* In the following, we have 0 <= i < A.length sv.
         Therefore, the next A.unsafe_set will be safe. *)
      A.iter_i (fun i s -> A.unsafe_set sv i (ignore_tag bg s)) sv;
      cons_wire_data_cell (bg, lazy (loop i))
    | Cons (Data _ as x, t) ->
      A.set sv i t;
      cons_wire_data_cell (x, lazy (loop (Nat.rem (Nat.succ i) n))) in
  lazy (loop _0_)
;;

let on_rails fv s =
  let n = A.length fv in
  let sv = fork_rails n s in
  (* Now, we have A.length sv = A.length fv = n.
     Therefore, in the following, with 0 <= i < n, the next A.unsafe_get will
     be safe. *)
  let fsv = A.init n (fun i -> (A.unsafe_get fv i) (A.unsafe_get sv i)) in
  join_rails fsv
;;

(* IO Functions *)

let rec map_wire_data f tl =
  let loop s =
    match get_cell s with
    | Nil -> tl (); nil_cell
    | Cons (h, t) -> cons_wire_data_cell (f h, map_wire_data f tl t) in
  fun s -> lazy (loop s)
;;

let rec iter_wire_data f s =
  match get_cell s with
  | Nil -> ()
  | Cons (h, t) -> f h; iter_wire_data f t
;;

let rec of_fun_wire_data =
  let loop f =
    match f () with
    | None -> nil_cell
    | Some x -> cons_wire_data_cell (x, of_fun_wire_data f) in
  fun f -> lazy (loop f)
;;

type 'a packet =
  | Dat of 'a wire_data
  | End
;;

let check_kind d1 d2 =
  match (d1, d2) with
  | (Begin_sub_stream, Begin_sub_stream)
  | (Data _, Data _)
  | (End_sub_stream, End_sub_stream) -> d2
  | ((Begin_sub_stream | Data _ | End_sub_stream),
     (Begin_sub_stream | Data _ | End_sub_stream)) -> assert false
;;

let of_to_io_channel (ic, oc) s =
  let send_recv x =
    IO.to_channel oc (Dat x) true;
    match IO.of_channel ic with
    | Dat y -> check_kind x y
    | End -> assert false in
  let tl = fun () ->
    IO.to_channel oc End true;
    match IO.of_channel ic with
    | Dat _ -> assert false
    | End -> () in
  map_wire_data send_recv tl s
;;

let to_io_channel s oc =
  let send x = IO.to_channel oc (Dat x) true in
  iter_wire_data send s;
  IO.to_channel oc End true
;;

let of_io_channel ic =
  let f () =
    match IO.of_channel ic with
    | Dat x -> Some x
    | End -> None in
  of_fun_wire_data f
;;

let stream_to_io_channel s oc =
  let send x =
    IO.to_channel oc (Dat x) true;
    match x with
    | Begin_sub_stream -> Begin_sub_stream
    | Data _  -> x
    | End_sub_stream -> End_sub_stream in
  let tl = fun () -> IO.to_channel oc End true in
  map_wire_data send tl s
;;

(* Debugging. *)

let print_stream data_printer ppf s =
  let rec print ppf s =
    match get_cell s with
    | Nil -> ()
    | Cons (Begin_sub_stream, t) ->
      Format.fprintf ppf "@;{@[<v 1>%a" print t
    | Cons (Data x, t) ->
      Format.fprintf ppf "@;Data %a%a" data_printer x print t
    | Cons (End_sub_stream, t) ->
      Format.fprintf ppf "@]@;}@;%a" print t in
  Format.fprintf ppf "@[<v> %a @]" print s
;;

let rec eq_stream s1 s2 =
  match (get_cell s1, get_cell s2) with
  | (Nil, Nil) -> true
  | (Nil, _)
  | (_, Nil) -> false
  | (Cons (Begin_sub_stream, t1), Cons (Begin_sub_stream, t2))
  | (Cons (End_sub_stream, t1), Cons (End_sub_stream, t2)) -> eq_stream t1 t2
  | (Cons (Data x1, t1), Cons (Data x2, t2)) -> x1 = x2 && eq_stream t1 t2
  | (Cons ((Begin_sub_stream | Data _ | End_sub_stream), _),
     Cons ((Begin_sub_stream | Data _ | End_sub_stream), _)) -> false
;;
