(***********************************************************************)
(*                                                                     *)
(*                                Sklml                                *)
(*                                                                     *)
(*                 The OCaml parallel skeleton system                  *)
(*                                                                     *)
(*                       EPI Serena, INRIA Paris                       *)
(*                                                                     *)
(*  Copyright (c) 2007-2017 INRIA.                                     *)
(*  Institut National de Recherche en Informatique et en Automatique.  *)
(*  All rights reserved. This file is distributed only by permission.  *)
(*                                                                     *)
(*  Quentin Carbonneaux                                                *)
(*  Francois Clement <Francois.Clement@inria.fr>                       *)
(*  Pierre Weis <Pierre.Weis@inria.fr>                                 *)
(*                                                                     *)
(***********************************************************************)

(* $Id: main.ml,v 1.61 2017/01/24 17:10:10 weis Exp $ *)

module Nat = Unatural;;
module L = Ulist;;
module L1 = Ulist1;;
module S = Ustring;;
module A = Uarray;;
module U = Uunix;;

let _0_ = Nat._0_;;

let nat_uunix_is_one_dead pids =
  U.is_one_dead (L.map Nat.to_int pids)
and nat_uunix_check_termination pid =
  U.check_termination (Nat.to_int pid)
;;

open Tools_common_types;;

let { tagged_debug = debug; _; } = Options.services;;

let debug fname = debug (Printf.sprintf "Main.%s" fname);;

let mk_cmd { Options.sklml_program_command_line = cmd; _; } master_node port =
  (* A.length cmd > 0, since it is built from a nonempty list.
     Therefore, the next A.unsafe_get will be safe. *)
  let debug = debug "mk_cmd" in
  let master_addr = Node.mk_addr master_node port in
  let ec = A.to_list cmd @ [ "-master"; master_addr; ] in
  debug (Printf.sprintf "command line = %S" (S.concat " " ec));
  let ec = A.of_list ec in
  { Batch.program = A.unsafe_get ec _0_; Batch.args = ec; }
;;

let sprint_master_slave_nodes_built n =
  Printf.sprintf "master and %d slave node(s) built" (Nat.to_int n)
and sprint_slave_batches_compiled n =
  Printf.sprintf "%d slave batch(es) compiled" (Nat.to_int n)
;;

let mk_batch
    ({ Options.timeout = tmout; _; } as options)
    (master_settings, slaves_settings) =
  let debug = debug "mk_batch" in
  let master_node = Node.mk_master options master_settings in
  let slave_nodes = L1.map (Node.mk_slave options) slaves_settings in
  let nslaves = L1.length slave_nodes in
  debug (sprint_master_slave_nodes_built nslaves);
  let master_batch = Node.compile_master options nslaves master_node in
  debug "master batch compiled";
  let slave_batches ic =
    let ic = Tio.tic_of_in_channel ic in
    let line =
      match Tio.input_line ic tmout with
      | Tio.Eof
      | Tio.Timeout -> ""
      | Tio.Line l -> l in
    match Parsers.parse_daemon_port line with
    | None -> Error.batch_add_fail_no_port line
    | Some port ->
      let b_slaves =
        Node.compile_slaves options master_node port slave_nodes in
      let run_skl_code pids =
        if nat_uunix_is_one_dead pids then
          Error.batch_add_fail_dead_slave () else
        begin match Tio.skip_n_input_lines ic nslaves tmout with
        | Tio.Eof -> Error.batch_add_fail_exited ()
        | Tio.Timeout -> Error.batch_add_fail_timeout ()
        | Tio.Line line ->
          begin match Parsers.parse_daemon_port line with
          | None -> Error.batch_add_fail_no_port line
          | Some port ->
            Batch.add_shell_task
              Batch.empty
              Error.batch_fail_client
              (mk_cmd options master_node port)
              Batch.Normal
          end
        end in
      Batch.add_control_task
        b_slaves
        Error.batch_fail_launch_client
        run_skl_code in
  debug (sprint_slave_batches_compiled nslaves);
  Batch.add_pipe_task
    Batch.empty
    Error.batch_fail_network
    master_batch
    slave_batches
;;

let check_for_batch_termination b =
  match Batch.eval_batch b stdout stderr with
  | Batch.Sklml_Ok pidl -> L.iter nat_uunix_check_termination pidl
  | Batch.Sklml_Error (pidl, _) -> Error.error_kill_jobs pidl
;;

let main () =
  let debug = debug "main" in
  let (options, sections) = Options.parse () in
  debug "command line parsed";
  try
    let all_settings = Parsers.parse_settings options sections in
    debug "configuration file parsed";
    let b = mk_batch options all_settings in
    debug "batch built";
    check_for_batch_termination b;
    debug "computation finished";
  with
  | Parsers.Parse_settings_error msg -> Error.error_parse_conf msg
  | Parsers.Expand_shell_command_error msg ->
    Error.error_expand_shell_command msg
;;

main ();;
