(***********************************************************************)
(*                                                                     *)
(*                                Sklml                                *)
(*                                                                     *)
(*                 The OCaml parallel skeleton system                  *)
(*                                                                     *)
(*                       EPI Serena, INRIA Paris                       *)
(*                                                                     *)
(*  Copyright (c) 2007-2017 INRIA.                                     *)
(*  Institut National de Recherche en Informatique et en Automatique.  *)
(*  All rights reserved. This file is distributed only by permission.  *)
(*                                                                     *)
(*  Quentin Carbonneaux                                                *)
(*  Francois Clement <Francois.Clement@inria.fr>                       *)
(*  Pierre Weis <Pierre.Weis@inria.fr>                                 *)
(*                                                                     *)
(***********************************************************************)

(* $Id: parsers.ml,v 1.24 2017/02/05 11:42:03 fclement Exp $ *)

module Nat = Unatural;;
module L = Ulist;;
module S = Ustring;;
module A = Uarray;;

let _0_ = Nat._0_
and _3_ = Nat._3_
;;

open Parser;;
open Parser_plus;;

type sections = string
and host = string
and shell = string
;;

type configuration_entry =
  | Nat of Nat.t
  | Str of string
and configuration_setting = string * configuration_entry
and configuration_settings = configuration_setting list
;;

let non_escape = any (L.map char (notc ['\\'; '"';]));;
                                   (* Helping Emacs '"' *)

let in_comm = any (L.map char (notc ['('; '*';]));;
let p_space = p_ignore p_space;;

let p_comment = (* parsing of nested comments *)
  let rec loop () =
    let rec loop_in () =
      many in_comm >>= fun _ ->
      (loop () >> loop_in ()) +++
      (p_word "*)" +++ eof >> return ()) +++
      ((char '(' +++ char '*') >> loop_in ()) in
    p_word "(*" >> loop_in () in
  loop ()
;;

let p_spaces = many (p_comment +++ p_space);;

let code_to_char l = Char.chr (to_num l);;

let p_string =
  let p_escape k =
    char '\\' >>
    (times _3_ p_digit >>= (fun l -> k (code_to_char l))) +++
    (char '"' >>= fun _ -> k '"') in
  let rec loop k =
    many non_escape >>= (fun x ->
    (p_escape (fun c -> loop (fun y -> k (x @ (c :: y))))) +++
    (char '"' >> k x)) in
  char '"' >> loop return
;;

let p_spaced p = p_spaces >> (p >>= (fun x -> p_spaces >> return x));;

let p_entry p =
  p_spaced (p_ident) >>= fun ident ->
  p_spaced (char '=') >> (p >>= fun x ->
  p_spaced (maybe (char ';')) >> (return (ident, x)))
;;

let p_entry =
  let p_value =
    (p_natural >>= (fun x -> return (Nat x))) +++
    (p_string >>= (fun x -> return (Str (S.gather x)))) in
  p_entry p_value
;;

let p_host = char '{' >> (many p_entry >>= fun l -> char '}' >> ( return l ));;

let parse_conf = full_parse (many (p_spaced p_host));;

let parse_daemon_port =
  let p_port =
    let p_not_colon = any (L.map char (notc [ ':'; ])) in
    let till_colon = many p_not_colon >> char ':' in
    (till_colon >> till_colon) >> p_natural in
  parse p_port
;;

let expand_shell_cmd host cmd =
  let p_in_word = any (L.map char (notc ('\'' :: '%' :: spaces))) in
  let p_in_quote = any (L.map char (notc [ '\''; '%'; ])) in
  let p_escape =
    (char 'h' >> return host) +++
    (char 'c' >> return cmd) +++
    (char '%' >> return "%") +++
    (char '\'' >> return "'") in
  let rec p_quote () =
    many p_in_quote >>= fun b ->
    let b = S.gather b in
    (char '%' >> (p_escape >>= fun esc ->
    p_quote () >>= fun e ->
    return (b ^ esc ^ e))) +++
    (char '\'' >> return b) in
  let p_end_of_word ret = p_ignore (many1 p_space) +++ eof >> return ret in
  let rec p_percent_word () =
    many p_in_word >>= fun b ->
    let b = S.gather b in
    (char '%' >> (p_escape >>= fun esc ->
    p_percent_word () >>= fun e ->
    return (b ^ esc ^ e))) +++
    (char '\'' >> (p_quote () >>= fun q ->
    p_percent_word () >>= fun e ->
    return (b ^ q ^ e))) +++
    p_end_of_word b in
  let p_cmd = many_till eof (p_percent_word ()) in
  parse p_cmd
;;

(* Handling of errors. *)

let truncate len msg = S.sub msg _0_ len ^ "...";;

let listlist2_of_listlist = function
  | [] -> ([], ([], []))
  | x :: [] -> (x, ([], []))
  | x :: y :: l -> (x, (y, l))
;;

exception Parse_settings_error of string;;

let parse_settings_error
    { Options.maximum_error_message_length = len; _; } msg =
  raise (Parse_settings_error (truncate len msg))
;;

let parse_settings options sections =
  match parse_conf sections with
  | None -> parse_settings_error options sections
  | Some (all_settings, "") -> listlist2_of_listlist all_settings
  | Some (_, s) -> parse_settings_error options s
;;

exception Expand_shell_command_error of string;;

let expand_shell_command_error
    { Options.maximum_error_message_length = len; _; } msg =
  raise (Expand_shell_command_error (truncate len msg))
;;

let expand_shell_command options host cmd shell =
  match expand_shell_cmd host cmd shell with
  | None -> expand_shell_command_error options shell
  | Some cmd -> A.of_list cmd
;;
