(***********************************************************************)
(*                                                                     *)
(*                                Sklml                                *)
(*                                                                     *)
(*                 The OCaml parallel skeleton system                  *)
(*                                                                     *)
(*                       EPI Serena, INRIA Paris                       *)
(*                                                                     *)
(*  Copyright (c) 2007-2017 INRIA.                                     *)
(*  Institut National de Recherche en Informatique et en Automatique.  *)
(*  All rights reserved. This file is distributed only by permission.  *)
(*                                                                     *)
(*  Quentin Carbonneaux                                                *)
(*  Francois Clement <Francois.Clement@inria.fr>                       *)
(*  Pierre Weis <Pierre.Weis@inria.fr>                                 *)
(*                                                                     *)
(***********************************************************************)

(* $Id: tio.ml,v 1.29 2017/01/04 11:58:30 fclement Exp $ *)

module Nat = Unatural;;
module Perv = Upervasives;;
module B = Ubytes;;

let _0_ = Nat._0_
and _1_ = Nat._1_
and _8192_ = Nat.unsafe_of_int 8192
;;

let nat_sys_max_string_length = Nat.unsafe_of_int Sys.max_string_length;;

let bts = B.to_string;;

let b_index_from_to s f t c =
  let len = B.length s in
  if len = _0_ then None else
  if t > len then None else
  let rec loop i =
    (* i is increasing from its initial value. *)
    if i >= t then None else
    (* Now, we have 0 <= i < t <= len.
       Therefore, the next B.unsafe_get will be safe. *)
    if B.unsafe_get s i = c then Some i else
    loop (Nat.succ i) in
  loop f
;;

let qual = Perv.qualify "Tio";;
let invalid_arg = Perv.invalid_arg_qual qual;;

let uunix_read fd buf ofs len =
  let n =
    Unix.read fd buf
      (Perv.offset_to_int ofs)
      (Perv.length_to_int len) in
  if n < 0 then invalid_arg "uunix_read" "negative number";
  Nat.of_int n
;;

open Tools_common_types;;

type tic =
  {
   mutable lr : Lrope.t;
   mutable buf : B.t;
   mutable ofs : offset;
   mutable len : length;
   descr : Unix.file_descr;
  }
;;

type 'a t =
  | Line of 'a
  | Eof
  | Timeout
and 'a tio = 'a t
;;

exception Tio_error of string;;
let tio_error m = raise (Tio_error m);;

let default_buf_size = _8192_;;

let tic_of_in_channel ic =
  let fd = Unix.descr_of_in_channel ic in
  {
   lr = Lrope.empty;
   buf = B.create default_buf_size;
   ofs = _0_;
   len = _0_;
   descr = fd;
  }
;;

let really_input ic t =
  let { descr = d; buf = buf; len = len; ofs = ofs; _ } = ic in
  let not_poll ic timeout =
    let (rl, _, _) = Unix.select [ic] [] [] (Nat.to_float timeout) in
    rl = [] in
  let blen = B.length buf in
  let olen = Nat.add len ofs in
  if olen = blen then Line () else
  if not_poll d t then Timeout else begin
    assert (olen <= blen);
    (* Now, we have olen <= blen.
       Therefore, the following Nat.sub will not fail. *)
    let len = uunix_read d buf olen (Nat.sub blen olen) in
    ic.len <- Nat.add ic.len len;
    if len = _0_ then Eof else
    Line ()
  end
;;

let pop_buffer_data ic s o l =
  let { lr = lr; buf = buf; len = len; ofs = ofs; _ } = ic in
  let lrlen = Lrope.length lr in
  if l <= lrlen then begin
    B.blit (Lrope.take_n lr l) _0_ s o l; (* sub optimal *)
    ic.lr <- Lrope.skip_n ic.lr l;
  end else begin
    (* Now, we have lrlen < l.
       Therefore, the next Nat.sub will not fail. *)
    B.blit (Lrope.take_n lr lrlen) _0_ s o lrlen;
    ic.lr <- Lrope.empty;
    let l = Nat.sub l lrlen in
    B.blit buf ofs s (Nat.add o lrlen) (min len l);
    if l >= len then begin
      ic.ofs <- _0_;
      ic.len <- _0_;
    end else begin
      (* Now, we have len < l.
         Therefore, the next Nat.sub will not fail. *)
      ic.len <- Nat.sub len l;
      ic.ofs <- Nat.add ic.ofs l;
    end
  end
;;

let lr_app l ic = Lrope.concat ic.lr (Lrope.rev_list_append Lrope.empty l);;

let input ic s o l t =
  if Nat.add o l > B.length s then invalid_arg "input" "" else
  let lrlen = Lrope.length ic.lr in
  if l <= Nat.add lrlen ic.len then begin
    pop_buffer_data ic s o l;
    Line l
  end else begin
    let rec loop acc acclen =
      match really_input ic t with
      | Line () ->
        if Nat.add acclen (Nat.add lrlen ic.len) >= l then begin
          ic.lr <- lr_app acc ic;
          pop_buffer_data ic s o l;
          Line l
        end else
          if Nat.add ic.ofs ic.len = B.length ic.buf then
            let (ofs, len) = (ic.ofs, ic.len) in
            ic.ofs <- _0_;
            ic.len <- _0_;
            loop (B.sub ic.buf ofs len :: acc) (Nat.add acclen len)
          else loop acc acclen
      | Eof ->
        let l_in = Nat.add acclen (Nat.add lrlen ic.len) in
        ic.lr <- lr_app acc ic;
        pop_buffer_data ic s o l;
        if l_in = _0_ then Eof else
        Line l_in
      | Timeout ->
        ic.lr <- lr_app acc ic;
        Timeout in
    loop [] _0_
  end
;;

let input_char ic t =
  let lrlen = Lrope.length ic.lr in
  if lrlen > _0_ then
    let c = Lrope.get_c ic.lr in
    ic.lr <- Lrope.skip_n ic.lr _1_;
    Line c
  else
    if ic.len > _0_ then
      (* Now, we have ic.len > 0.
         Therefore, the next Nat.pred will not fail. *)
      let c = B.unsafe_get ic.buf ic.ofs in
      ic.ofs <- Nat.succ ic.ofs;
      ic.len <- Nat.pred ic.len;
      Line c
    else begin
      if ic.ofs = B.length ic.buf then ic.ofs <- _0_;
      assert (ic.len > _0_);
      (* Now, we have ic.len > 0.
         Therefore, the next Nat.pred will not fail. *)
      match really_input ic t with
      | Line () ->
        let c = B.unsafe_get ic.buf ic.ofs in
        ic.ofs <- Nat.succ ic.ofs;
        ic.len <- Nat.pred ic.len;
        Line c
      | Eof
      | Timeout as err -> err
    end
;;

let input_line ic t =
  let line_too_long () = tio_error "line length exceeds max_string_length" in
  let rec loop acc =
    match really_input ic t with
    | Line () ->
      let idx =
        b_index_from_to ic.buf ic.ofs (Nat.add ic.ofs ic.len) '\n' in
      begin match idx with
      | None ->
        if Nat.add ic.ofs ic.len <> B.length ic.buf then loop acc else
        let (len, ofs) = (ic.len, ic.ofs) in
        ic.ofs <- _0_;
        ic.len <- _0_;
        loop (B.sub ic.buf ofs len :: acc)
      | Some i ->
        ic.lr <- lr_app acc ic;
        (* tlen = Lrope.length ic.lr + i + 1. *)
        let tlen = Nat.add (Lrope.length ic.lr) (Nat.succ i) in
        assert (tlen >= ic.ofs);
        (* Now, we have tlen >= ic.ofs.
           Therefore, the next Nat.sub will not fail. *)
        let tlen = Nat.sub tlen ic.ofs in
        if tlen > nat_sys_max_string_length then line_too_long () else
        let s = B.create tlen in
        pop_buffer_data ic s _0_ tlen;
        Line (bts s)
      end
    | Eof ->
      let tlen = Nat.add (Lrope.length ic.lr) ic.len in
      if tlen = _0_ then Eof else
      if tlen > nat_sys_max_string_length then line_too_long () else
      let s = B.create tlen in
      pop_buffer_data ic s _0_ tlen;
      Line (bts s)
    | Timeout ->
      ic.lr <- lr_app acc ic;
      Timeout in
  let lrlen = Lrope.length ic.lr in
  let ilr_ibuf =
    (Lrope.index ic.lr '\n',
     b_index_from_to ic.buf ic.ofs (Nat.add ic.ofs ic.len) '\n') in
  match ilr_ibuf with
  | (None, None) -> loop []
  | (None, Some ibuf) ->
    (* tlen = lrlen + ibuf + 1. *)
    let tlen = Nat.add lrlen (Nat.succ ibuf) in
    assert (tlen >= ic.ofs);
    (* Now, we have tlen >= ic.ofs.
       Therefore, the next Nat.sub will not fail. *)
    let tlen = Nat.sub tlen ic.ofs in
    if tlen > nat_sys_max_string_length then line_too_long () else
    let s = B.create tlen in
    pop_buffer_data ic s _0_ tlen;
    Line (bts s)
  | (Some ilr, _) ->
    let tlen = Nat.succ ilr in
    if tlen > nat_sys_max_string_length then line_too_long () else
    let s = Lrope.take_n ic.lr tlen in
    pop_buffer_data ic s _0_ tlen;
    Line (bts s)
;;

let rec skip_n_input_lines ic n t =
  if n = _0_ then input_line ic t else
  (* Now, we have n > 0.
     Therefore, the next Nat.pred will not fail. *)
  match input_line ic t with
  | Line _ -> skip_n_input_lines ic (Nat.pred n) t
  | Eof
  | Timeout as err -> err
;;
