(***********************************************************************)
(*                                                                     *)
(*                                Sklml                                *)
(*                                                                     *)
(*                 The OCaml parallel skeleton system                  *)
(*                                                                     *)
(*                       EPI Serena, INRIA Paris                       *)
(*                                                                     *)
(*  Copyright (c) 2007-2017 INRIA.                                     *)
(*  Institut National de Recherche en Informatique et en Automatique.  *)
(*  All rights reserved. This file is distributed only by permission.  *)
(*                                                                     *)
(*  Quentin Carbonneaux                                                *)
(*  Francois Clement <Francois.Clement@inria.fr>                       *)
(*  Pierre Weis <Pierre.Weis@inria.fr>                                 *)
(*                                                                     *)
(***********************************************************************)

(* $Id: tools_options.ml,v 1.27 2017/01/04 11:58:30 fclement Exp $ *)

module Nat = Unatural;;
module L = Ulist;;
module A = Uarray;;
module Ar = Uaref;;
module Ok = Uoption_key;;
module Co = Ucommand_option;;
module D = Udebug;;
module Pic = Upph_id_card;;

let _0_ = Nat._0_
and _1_ = Nat._1_
;;

open Tools_common_types;;

(* Command-line options. *)

let print_and_exit msg = fun () -> Printf.eprintf "%s\n" msg; exit 0;;

let semantics_mode_option_name = "semantics mode"
and sklml_install_root_dir_option_name = "Sklml installation root directory"
;;

let semantics_mode_symbols = [ "seq"; "par"; ];;

let default_semantics_mode = Seq
and default_compilation_mode = Byt
and default_sklml_install_root_dir =
  Sklml_install_data.default_install_root_dir
;;

let other_compilation_mode =
  Sklml_mode.other_compilation_mode default_compilation_mode
;;

let { Ar.get_list = get_files_to_compile;
      add_item = add_file_to_compile; _; } =
  Ar.make_aref_list []

and (get_semantics_mode, set_semantics_mode) =
  let { Ar.get; set; } =
    Ar.make_aref_single
      semantics_mode_option_name default_semantics_mode in
  let set s = set (Sklml_mode.semantics_mode_of_string s) in
  (get, set)

and (get_compilation_mode, set_native_code_compilation_mode) =
  let { Ar.get_flag; set_flag; _; } = Ar.make_flag () in
  let get_flag () =
    if get_flag () then other_compilation_mode else
    default_compilation_mode in
  (get_flag, set_flag)

and { Ar.get = get_sklml_install_root_dir;
      set = set_sklml_install_root_dir; } =
  Ar.make_aref_single
    sklml_install_root_dir_option_name
    default_sklml_install_root_dir
;;

let get_sklml_lib_root_dir () =
  Sklml_path.get_sklml_lib_root_dir (get_sklml_install_root_dir ())

and get_sklml_pph_tekno_dir () =
  Sklml_path.get_sklml_pph_tekno_dir (get_sklml_install_root_dir ())
;;

let semantics_mode_ok = Ok.of_string "-mode"
and compilation_mode_ok = Ok.of_string "-O"
and sklml_install_root_dir_ok = Ok.of_string "-sklml-install-root-dir"
and sklml_lib_root_dir_ok = Ok.of_string "-where"
and sklml_pph_tekno_dir_ok = Ok.of_string "-pph-tekno"
;;

let semantics_mode_rco =
  {
   Co.key = semantics_mode_ok;
   spec = Arg.Symbol (semantics_mode_symbols, set_semantics_mode);
   doc =
     Printf.sprintf
       "\n\tSet the %s. Default is %S."
       semantics_mode_option_name
       (Sklml_mode.string_of_semantics_mode (get_semantics_mode ()));
  }

and compilation_mode_rco =
  {
   Co.key = compilation_mode_ok;
   spec = Arg.Unit set_native_code_compilation_mode;
   doc =
     Printf.sprintf
       "\n\tInvoke %s.\n\
       \tDefault is to invoke %s."
       (Sklml_mode.doc_string_of_compilation_mode other_compilation_mode)
       (Sklml_mode.doc_string_of_compilation_mode default_compilation_mode);
  }

and sklml_install_root_dir_rco =
  {
   Co.key = sklml_install_root_dir_ok;
   spec = Arg.String set_sklml_install_root_dir;
   doc =
     Printf.sprintf
       "<path>\n\
       \tSet the installation root directory of the Sklml system to <path>.\n\
       \tDefault is %S."
       (get_sklml_install_root_dir ());
  }

and sklml_lib_root_dir_rco =
  {
   Co.key = sklml_lib_root_dir_ok;
   spec = Arg.Unit (print_and_exit (get_sklml_lib_root_dir ()));
   doc = Printf.sprintf "\n\tPrint the location of Sklml library and exit.";
  }

and sklml_pph_tekno_dir_rco =
  {
   Co.key = sklml_pph_tekno_dir_ok;
   spec = Arg.Unit (print_and_exit (get_sklml_pph_tekno_dir ()));
   doc = Printf.sprintf "\n\tPrint the location of Sklml Pph tekno and exit.";
  }
;;

let set_command_line_options
    command_kind command_suffix
    oks rcos
    doc_mandatory_arguments =

  let project_name = Sklml_version.pph_id_card.Pic.project_name in

  let sklml_command_kind = Printf.sprintf "%s %s" project_name command_kind
  and ocaml_command_kind = Printf.sprintf "OCaml %s" command_kind in

  let command_title = Printf.sprintf "%s%s" project_name command_suffix in

  let phase = Co.Phase sklml_command_kind in

  let add_options () =
    Co.application_add_all_rco rcos;
    Co.phase_add_all phase oks in

  add_options ();

  Pic.add_app_version_options Sklml_version.pph_id_card None;

  Co.set_anon_fun add_file_to_compile;

  (* Sys.argv is never empty.
     Therefore, the next A.unsafe_get will be safe. *)
  let cmd_name = A.unsafe_get Sys.argv _0_ in
  Co.set_usage_msg
    (Format.sprintf
       "@[<v>\
       @ \
       %s -- %s@ \
       @ \
       Usage:@ \
       @ \
       %s [<option>...] %s<file>...@ \
       @ \
       Arguments are (in addition to those of the %s):\
       @]@."
       cmd_name sklml_command_kind
       cmd_name doc_mandatory_arguments
       ocaml_command_kind);

  (command_title, phase)
;;

(* Report/debug flags and warning/error messages. *)

let add_services kd sname =
  let pf = D.application_printing_facilities kd in
  D.warn_error_app_clear ();
  Co.phase_all_services ();
  let report msg = pf.D.report (Printf.sprintf "%s: %s" sname msg)
  and debug = pf.D.debug (Some sname)
  and error msg ier =
    pf.D.error (Some sname) msg;
    exit (Nat.to_int ier) in
  { report; debug; error; }
;;

let add_tagged_services sname =
  let kd =
    { Sklml_version.option_key_data_no_prefix with Co.name = sname; } in
  let pf = D.application_printing_facilities_with_static_tag kd in
  Co.phase_all_services ();
  let tagged_report = pf.D.report
  and tagged_report_action = pf.D.report_action
  and tagged_debug fname = pf.D.debug (Some fname)
  and tagged_debug_action fname = pf.D.debug_action (Some fname)
  and tagged_error msg = pf.D.error None msg; exit 1 in
  let tagged_warning msg =
    pf.D.warning None msg;
    if pf.D.get_warn_error () then
      tagged_error "Warn_error enabled, stopping here" in
  { tagged_report; tagged_report_action;
    tagged_debug; tagged_debug_action;
    tagged_warning; tagged_error; }
;;

(* Parsing the command line. *)

(* We need here to collect non-Sklml options, that are supposed to be parse by
 the corresponding wrapped OCaml command, hence the use of parse*_argv
 functions, for which it is necessary to catch exceptions. *)
let parse_aux phase argv =
  let argv = Co.parse_all_services_argv argv in
  Co.parse_phase_argv phase argv
;;

let get_ocaml_options argv =
  L.tl (A.to_list argv) @ get_files_to_compile ()
;;

let parse { error; _; } phase get_options =
  let argv = Co.arg_catch (parse_aux phase) Sys.argv in
  try get_options argv with
  | Failure msg -> error msg _1_
;;
